# -*- coding: utf-8 -*-
"""
Main module.
"""
from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "tum-exam-scripts"
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound
__author__ = "Patrick Stoeckle"
__copyright__ = "Patrick Stoeckle"
__license__ = "mit"



"""
PDF commands.
"""
from pathlib import Path
from typing import List, Optional

from click import echo
from click.exceptions import Exit

from tum_exam_scripts.logic.pdf_printing import (
    send_attendee_list_internal,
    send_pdf_files,
)
from tum_exam_scripts.shared import DRIVER_OPTION
from tum_exam_scripts.utils.command import call_command, confirm_printing_rights
from typer import Argument, Option, Typer

app = Typer()


@app.callback()
def _call_back() -> None:
    """
    Subgroup with the PDF printing commands.
    """


@app.command()
def send_all_booklets(
    driver_name: str = DRIVER_OPTION,
    input_directory: Path = Argument(
        ".",
        exists=True,
        resolve_path=True,
        help="The directory with the exams from the TUMExam website.",
        file_okay=False,
    ),
    batch_size: Optional[int] = Option(
        None,
        "--batch-size",
        "-b",
        help="If you add a batch size, the process will stop after so many exams and wait for you to continue."
        "You can you this so start all jobs on a printer, then send the next batch, and start these exams on another printer.",
    ),
) -> None:
    """
    Send all booklets to the printing server.

    Example:
        tum-exam-scripts send-all-booklets /path/to/exams/
    """
    if batch_size is not None:
        if batch_size < 2:
            echo(f"{batch_size} is not a valid batch size!")
            raise Exit(1)
    confirm_printing_rights()
    pdf_files = sorted(input_directory.glob("*-book.pdf"))
    if len(pdf_files) == 0:
        echo(f"We did not find any booklets. Please check {input_directory}")
        raise Exit(1)
    echo(f"We found {len(pdf_files)} booklets.")
    send_pdf_files(driver_name, pdf_files, batch_size)


@app.command()
def send_specific_booklets(
    pdf_file: List[Path] = Argument(
        None,
        exists=True,
        resolve_path=True,
        help="The directory with the exams from the TUMExam website.",
        dir_okay=False,
    ),
    driver_name: str = DRIVER_OPTION,
) -> None:
    """
    Send only specific PDFs to the server. You can pass multiple files.

    Example:
        tum-exam-scripts send-specific-booklets /path/to/E0007-book.pdf /path/to/E0009-book.pdf
    """
    confirm_printing_rights()
    send_pdf_files(driver_name, pdf_file)


@app.command()
def send_attendee_list(
    attend_list: Path = Argument(
        "attendeelist.pdf",
        exists=True,
        resolve_path=True,
        help="The attendee list from the TUMExam endterm_lists folder.",
        dir_okay=False,
    ),
    driver_name: str = DRIVER_OPTION,
) -> None:
    """
    Send the attendee list to the server.

    Example:
        tum-exam-scripts send-attendee-list /path/to/attendeelist.pdf
    """
    confirm_printing_rights()
    send_attendee_list_internal(attend_list, driver_name)


@app.command()
def send_seat_plan(
    seat_plan: Path = Argument(
        "seatplan-a3.pdf",
        exists=True,
        resolve_path=True,
        help="The seat plan in A3 from the TUMExam endterm_lists folder.",
        dir_okay=False,
    ),
    driver_name: str = DRIVER_OPTION,
    versions: int = Option(
        3, "--number-of-copies", "-n", help="The number of copies you want to print."
    ),
) -> None:
    """
    Print the seat plans in A3. You have to put them at the doors of the lecture hall.
    """
    confirm_printing_rights()
    echo(f"Sending document {seat_plan} to the printing server ...")
    current_command = [
        "lp",
        "-d" + driver_name,
        "-n",
        str(versions),
        "-o",
        "PageSize=A3",
        "-o",
        "JCLBanner=False",
        "-o",
        "JCLColorCorrection=BlackWhite",
        "-o",
        "Duplex=None",
        "-o",
        "JCLPrintQuality=Enhanced",
        str(seat_plan),
    ]
    call_command(seat_plan, current_command)


@app.command()
def send_room_layout(
    room_plan: Path = Argument(
        "roomplan.pdf",
        exists=True,
        resolve_path=True,
        help="The room plan in A3 from the TUMExam endterm_lists folder.",
        dir_okay=False,
    ),
    driver_name: str = DRIVER_OPTION,
    versions: int = Option(
        3, "--number-of-copies", "-n", help="The number of copies you want to print."
    ),
) -> None:
    """
    Print the room plans in A3. You have to put them at the doors of the lecture hall.
    """
    confirm_printing_rights()
    echo(f"Sending document {room_plan} to the printing server ...")
    current_command = [
        "lp",
        "-d" + driver_name,
        "-n",
        str(versions),
        "-o",
        "PageSize=A3",
        "-o",
        "JCLBanner=False",
        "-o",
        "JCLColorCorrection=PressMatch",
        "-o",
        "Duplex=None",
        "-o",
        "JCLPrintQuality=Enhanced",
        str(room_plan),
    ]
    call_command(room_plan, current_command)



"""
Utils.
"""
import subprocess  # NOTE: Keep for mock/testing
from logging import getLogger
from pathlib import Path
from subprocess import PIPE, Popen
from typing import List, Sequence

from click import echo, style
from click.exceptions import Exit

import typer  # NOTE: Keep for mock/testing
from typer.colors import RED

_LOGGER = getLogger(__name__)


def error_echo(s: str) -> None:
    """

    :param s:
    :return:
    """
    echo(style(s, fg=RED), err=True)


def call_command(current_file: Path, current_command: Sequence[str]) -> None:
    """
    Call a command and exits on failure.
    :param current_file:
    :param current_command:
    :return:
    """
    _LOGGER.info("Calling ...")
    _LOGGER.info(" ".join(current_command))
    _LOGGER.info("Done!")
    res = subprocess.check_call(current_command)
    if res != 0:
        error_echo(f"Something went wrong when sending {current_file} to the server")
        error_echo(f"Please open a shell and call {' '.join(current_command)}")
        raise Exit(1)


def confirm_printing_rights() -> None:
    """
    Ask the user whether they have enabled the printing.
    """
    if not typer.confirm(
        "Did you enable printing from your PC via https://ucentral.in.tum.de/cgi-bin/printman.cgi ?"
    ):
        echo("Please enable printing first!")
        raise Exit


def sudo_call(command: List[str], password: str) -> None:
    """
    Sudo call.
    """
    changed_command = ["sudo", "-S"] + command
    _LOGGER.info("Calling ...")
    _LOGGER.info(" ".join(changed_command))
    _LOGGER.info("Done!")
    proc = Popen(
        changed_command,
        stdin=PIPE,
        stdout=PIPE,
        stderr=PIPE,
    )
    proc.communicate(password.encode())
    if proc.returncode != 0:
        error_echo("Installation went wrong.")
        error_echo(f"Please open a shell and call 'sudo {' '.join(command)}'")
        raise Exit(1)



"""
Main.
"""
from logging import INFO, basicConfig, getLogger
from typing import Optional

from tum_exam_scripts import __version__
from tum_exam_scripts.enums import Browser
from tum_exam_scripts.logic.pdf_printing import install_linux_driver_internal
from tum_exam_scripts.pdf_commands import app as pdf_commands_app
from tum_exam_scripts.shared import DRIVER_OPTION
from tum_exam_scripts.utils.password_handling import (
    get_password_from_keyring,
    store_password,
)
from tum_exam_scripts.utils.website import open_website_internal
from typer import Argument, Exit, Option, Typer, echo

_USER_ARGUMENT = Argument(
    None,
    help="The username for your informatics account, i.e., the first letters of your lastname.",
)
_LOGGER = getLogger(__name__)
basicConfig(
    format="%(levelname)s: %(asctime)s: %(name)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=INFO,
    filename="tum-exam-scripts.log",
    filemode="a",
)


def _version_callback(value: bool) -> None:
    if value:
        echo(f"tum-exam-scripts {__version__}")
        raise Exit()


app = Typer()


@app.callback()
def _call_back(
    _: bool = Option(
        None,
        "--version",
        is_flag=True,
        callback=_version_callback,
        expose_value=False,
        is_eager=True,
        help="Version",
    )
) -> None:
    """
    A collection of useful commands to print TUMExams. You can find the source code under https://gitlab.lrz.de/i4/software/tum-exam-scripts
    """


@app.command()
def install_linux_driver(
    driver_name: str = DRIVER_OPTION,
    user_password: str = Option(
        None,
        "--password",
        "-p",
        help="Your user password. NOTE: The user should have 'sudo' privileges.",
        hide_input=True,
        prompt=True,
    ),
) -> None:
    """
    This snippet downloads the Linux driver for the printers and makes them available under $driver_name
    This is needed as the macOS driver cannot handle the booklets.
    Please change the command on mac for printing the exams from `-dfollowme` to `-dfollowmepdd`!!!
    """
    install_linux_driver_internal(driver_name, user_password)


@app.command()
def store_password_in_password_manager(
    password: str = Option(
        None,
        "--password",
        "-p",
        prompt=True,
        hide_input=True,
        help="The password for your informatics account",
    ),
    user_name: str = _USER_ARGUMENT,
    force: bool = Option(
        False,
        "--force",
        "-f",
        is_flag=True,
        help="If true, we will overwrite existing passwords.",
    ),
) -> None:
    """
    Stores the password in the password manager.
    """
    store_password(force, password, user_name)


@app.command()
def open_printing_page(
    user_name: str = _USER_ARGUMENT,
    browser: Browser = Option(
        "firefox", "--browser", "-b", help="The browser to start."
    ),
    password: Optional[str] = Option(
        None,
        "--password",
        "-p",
        help="The password for your informatics account. If you do not pass a password, we will use the password stored in the password manager.",
    ),
) -> None:
    """
    Open the page we need to send the PDFs to the FollowMe printer.

    """
    if password is None:
        password = get_password_from_keyring(user_name)
    open_website_internal(user_name, password, browser)


app.add_typer(pdf_commands_app, name="pdf")

if __name__ == "__main__":
    app()



"""
PDF utils.
"""
from logging import getLogger
from os import remove
from os.path import getsize
from pathlib import Path
from tempfile import gettempdir
from typing import List, Optional
from urllib.request import urlretrieve

from click import echo, pause
from click.exceptions import Exit
from tqdm import tqdm

from tum_exam_scripts.utils.command import call_command, error_echo, sudo_call

_LOGGER = getLogger(__name__)


def is_full_pdf(current_file: Path) -> bool:
    """
    Check whether a file is a valid PDF.
    :param current_file:
    :return:
    """
    size = getsize(current_file)
    if size < 1024:
        return False
    with current_file.open("rb") as fin:
        # start content
        fin.seek(0)
        start_content = fin.read(1024).decode("ascii", "ignore")
        fin.seek(-1024, 2)
        end_content = fin.read().decode("ascii", "ignore")
    start_flag = False
    # %PDF
    if start_content.count("%PDF") > 0:
        start_flag = True

    if end_content.count("%%EOF") and start_flag > 0:
        return True
    eof: str = bytes([0]).decode("ascii")
    if end_content.endswith(eof) and start_flag:
        return True
    return False


def send_pdf_files(
    driver_name: str, pdf_files: List[Path], batch_size: Optional[int] = None
) -> None:
    """
    Send all PDF files to the server.
    :param batch_size:
    :param driver_name:
    :param pdf_files:
    :return:
    """
    echo("Check whether PDFs are corrupt")
    for pdf_file in tqdm(pdf_files):
        if not is_full_pdf(pdf_file):
            error_echo(f"The PDF file {pdf_file} is not a valid PDF.")
            raise Exit(1)
    batch_no = 0
    for i, pdf_file in enumerate(tqdm(pdf_files)):
        echo(f"Sending document {pdf_file} to the printing server ...")
        current_command = [
            "lp",
            "-d" + driver_name,
            "-o",
            "PageSize=A3",
            "-o",
            "JCLBanner=False",
            "-o",
            "JCLColorCorrection=BlackWhite",
            "-o",
            "Duplex=DuplexNoTumble",
            "-o",
            "XRFold=BiFoldStaple",
            "-o",
            "landscape",
            "-o",
            "JCLPrintQuality=Enhanced",
            str(pdf_file),
        ]
        call_command(pdf_file, current_command)
        if batch_size is not None and ((i + 1) % batch_size) == 0:
            pause(f"We finished batch {batch_no}")
            batch_no += 1

    echo("Done!")


def install_linux_driver_internal(driver_name: str, user_password: str) -> None:
    tempdir = Path(gettempdir())
    local_file = tempdir.joinpath("x2UNIV.ppd")
    _LOGGER.info("Download PPD file")
    urlretrieve(
        "https://wiki.in.tum.de/foswiki/pub/Informatik/Benutzerwiki/XeroxDrucker/x2UNIV.ppd",
        str(local_file),
    )
    _LOGGER.info("Success!")
    sudo_call(
        [
            "lpadmin",
            "-E",
            "-p",
            driver_name,
            "-v",
            "ipps://print.in.tum.de/printers/followme",
            "-P",
            str(local_file),
            "-D",
            "Xerox-Followme",
            "-L",
            "TUM",
        ],
        user_password,
    )
    echo("The Linux driver was successfully installed!")
    remove(local_file)
    sudo_call(["cupsenable", driver_name], user_password)
    sudo_call(["cupsaccept", driver_name], user_password)
    echo(f"The printing service is available under {driver_name}")


def send_attendee_list_internal(attend_list: Path, driver_name: str) -> None:
    echo(f"Sending document {attend_list} to the printing server ...")
    current_command = [
        "lp",
        "-d" + driver_name,
        "-o",
        "PageSize=A4",
        "-o",
        "JCLBanner=False",
        "-o",
        "JCLColorCorrection=PressMatch",
        "-o",
        "Duplex=None",
        "-o",
        "JCLPrintQuality=Enhanced",
        "-o",
        "InputSlot=ManualFeed",
        "-o",
        "MediaType=Labels",
        str(attend_list),
    ]
    call_command(attend_list, current_command)



"""
Password handling
"""
from logging import getLogger
from sys import platform

from keyring import get_password, set_keyring, set_password
from keyring.backends.macOS import Keyring as macos_Keyring
from keyring.backends.Windows import WinVaultKeyring

from typer import Exit, echo

SERVICE_NAME = "tum-exam-scripts"

_LOGGER = getLogger(__name__)


def store_password(force: bool, password: str, user_name: str) -> None:
    """

    :param force:
    :param password:
    :param user_name:
    :return:
    """
    _set_keyring()
    current_password = get_password(SERVICE_NAME, user_name)
    if current_password is not None:
        echo(f"There is already a password stored for {user_name}")
        if not force:
            return
    set_password(SERVICE_NAME, user_name, password)
    echo("We stored the password!")


def get_password_from_keyring(user_name: str) -> str:
    """

    :param user_name:
    :return:
    """
    _set_keyring()
    p = get_password(SERVICE_NAME, user_name)
    if p is None:
        echo(f"No password stored for {user_name}")
        raise Exit(1)
    return p


def _set_keyring() -> None:
    if platform == "linux" or platform == "linux2":
        echo("Please configure the right keyring!!!")
    elif platform == "darwin":
        set_keyring(macos_Keyring())
    elif platform == "win32":
        set_keyring(WinVaultKeyring())



"""
Anything related to Selenium and the printing page.
"""
from click import echo
from click.exceptions import Exit

from helium import (
    Button,
    click,
    refresh,
    start_chrome,
    start_firefox,
    wait_until,
    write,
)
from tum_exam_scripts.enums import Browser


def open_website_internal(user_name: str, password: str, browser: Browser) -> None:
    """
    Use selenium to open the printing page.
    """
    url = "https://ucentral.in.tum.de/cgi-bin/index.cgi"
    if browser == Browser.CHROME:
        start_chrome(url)
    elif browser == Browser.FIREFOX:
        start_firefox(url)
    else:
        echo(f"Browser {browser} is not supported!")
        raise Exit(1)
    click("Login")
    write(user_name, "User:")
    write(password, "Password:")
    click("Login")
    wait_until(Button("Logout").exists, timeout_secs=5)
    click("Xerox Printing")
    wait_until(Button("Laden").exists, timeout_secs=5)
    refresh()
    try:
        wait_until(
            Button("Diesen Rechner zum Drucken freischalten").is_enabled, timeout_secs=5
        )
        click("Diesen Rechner zum Drucken freischalten")
        refresh()
        click("Diesen Rechner zum Drucken freischalten")
    except LookupError:
        echo("We can already print from this machine")
    echo("NOTE: Keep the browser window open!")
